==> qmpdclient-1.0.6.2.tar.gz: qmpdclient-1.0.6.2/src/playlistview.cpp <==
/*
 * QMPDClient - An MPD client written in Qt 4.
 * Copyright (C) 2005-2006 Hvard Tautra Knutsen <havtknut@tihlde.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"
#include "jumptosongdialog.h"
#include "metainfodialog.h"
#include "mpd.h"
#include "mpddirectory.h"
#include "mpdsong.h"
#include "mpdsongmodel.h"
#include "playlistview.h"
#include <QDragMoveEvent>
#include <QInputDialog>
#include <QMenu>
#include <QPainter>
#include <QScrollBar>

#define DROPEND -7
#define DONTDROP -10

PlaylistView::PlaylistView(QWidget *parent) : MPDSongView(parent, HeaderView::PLAYLISTVIEW, false),
		m_dropPos(DONTDROP),
m_hidden(false) {
	m_model->setExtraFlags(Qt::ItemIsDropEnabled);
	setDropIndicatorShown(true);

	// Disconnect the default action from super class
	disconnect(this, SIGNAL(activated(const QModelIndex &)), this, SLOT(indexActivated(const QModelIndex &)));

	// playlist menu
	QAction *jumpToAction = m_menu->addAction(QPixmap(":/icons/crystal/goto.png"), tr("J&ump to song..."), this, SLOT(jumpToSong()), tr("Ctrl+U", "This is the shortcut for 'Jump to song in playlist'"));

	addAction(jumpToAction);
	m_menu->addSeparator();
	m_removeAction = m_menu->addAction(QPixmap(":/icons/remove.png"), tr("&Remove"), this, SLOT(removeItems()));
	m_removeAction->setShortcut(Qt::Key_Delete);
	m_removeAction->setShortcutContext(Qt::WidgetShortcut);
	addAction(m_removeAction);
	m_cropAction = m_menu->addAction(QPixmap(":/icons/crop.png"), tr("&Crop playlist"), this, SLOT(cropItems()));
	m_menu->addAction(QPixmap(":/icons/removeall.png"), tr("Remove &all"), m_mpd, SLOT(clearPlaylist()));
	m_menu->addSeparator();
	m_randomAction = m_menu->addAction(QPixmap(":/icons/amarok/random.png"), tr("Ra&ndom play"));
	m_repeatAction = m_menu->addAction(QPixmap(":/icons/amarok/repeat.png"), tr("Re&peat play"));
	m_menu->addAction(QPixmap(":/icons/amarok/shuffle.png"), tr("S&huffle playlist"), m_mpd, SLOT(shufflePlaylist()));
	m_menu->addSeparator();
	m_menu->addAction(QPixmap(":/icons/crystal/saveas.png"), tr("&Save playlist..."), this, SLOT(savePlaylist()));
	m_informationAction = m_menu->addAction(QPixmap(":/icons/crystal/info.png"), tr("&Information..."), this, SLOT(information()));

	m_randomAction->setCheckable(true);
	m_repeatAction->setCheckable(true);

	// signals and slots connections
	connect(m_mpd, SIGNAL(playingSongUpdated(MPDSong *)), this, SLOT(setSong(MPDSong *)));
	connect(m_mpd, SIGNAL(playlistUpdated(MPDSongList)), this, SLOT(setPlaylist(MPDSongList)));
	connect(m_mpd, SIGNAL(randomUpdated(bool)), m_randomAction, SLOT(setChecked(bool)));
	connect(m_mpd, SIGNAL(repeatUpdated(bool)), m_repeatAction, SLOT(setChecked(bool)));
	connect(m_randomAction, SIGNAL(toggled(bool)), m_mpd, SLOT(setRandom(bool)));
	connect(m_repeatAction, SIGNAL(toggled(bool)), m_mpd, SLOT(setRepeat(bool)));
	connect(this, SIGNAL(activated(const QModelIndex &)), this, SLOT(indexActivated(const QModelIndex &)));

	selectionChanged();
}

void PlaylistView::selectionChanged(const QItemSelection &s, const QItemSelection &u) {
	MPDSongView::selectionChanged(s, u);
	bool t = selectedIndexes().count() > 0;
	m_cropAction->setEnabled(t);
	m_removeAction->setEnabled(t);
	emit toggleActions(t);
}

void PlaylistView::indexActivated(const QModelIndex &index) {
	foreach(MPDSong *s, m_model->songs(index)) {
		m_mpd->jumpPlaylistSong(s);
		break;
	}
}

void PlaylistView::cropItems() {
	m_mpd->removeSongs(m_model->unSelectedSongs(selectedIndexes()));
}

void PlaylistView::removeItems() {
	QModelIndexList selected = selectedIndexes();
	if (selected.isEmpty())
		return;
	m_mpd->removeSongs(m_model->songs(selected));
}

void PlaylistView::setPlaylist(MPDSongList songs) {
	QModelIndexList selected = selectedIndexes();
	QModelIndex last = selected.isEmpty() ? QModelIndex() : selected.last();
	m_model->setSongs(songs);
	setCurrentIndex(last);
	selectionChanged();
}

void PlaylistView::setSong(MPDSong *s) {
	m_model->setPlaying(s);
	// scrollTo is not nice if you are messing with the playlist.
	// Scroll only if no items selected
	if (s && m_config->scrollToPlaying()) {
		// Find first valid index to work on
		// TODO, when Qt 4.2 ships, we can hopefully do this:
		//scrollTo(m_model->indexId(s->id()), QAbstractItemView::PositionAtCenter);
		m_hidden = true;
		scrollTo();
	}
}

void PlaylistView::hideEvent(QHideEvent *e) {
	MPDSongView::hideEvent(e);
	m_hidden = true;
}
void PlaylistView::paintEvent(QPaintEvent *e) {
	MPDSongView::paintEvent(e);
	scrollTo();
}

void PlaylistView::scrollTo(QModelIndex index) {
	if (!index.isValid()) {
		index = m_model->indexId(m_model->playingId());
		if (!index.isValid())
			return;
	}

	if (isIndexHidden(index)) {
		for (int i=0; i < 12; i++) {
			index = m_model->index(index.row(), i);
			if (!isIndexHidden(index))
				break;
		}
	}

	QRect area = viewport()->rect();
	QRect rect = visualRect(index);

	QScrollBar *sb = verticalScrollBar();
	MPDSongList songs = m_model->songs(index);

	if (songs.isEmpty() || rect.isEmpty() || !m_hidden || sb->isHidden())
		return;

	int indexOfSong = m_model->songs().indexOf(songs.first());
	int heightAllItems = rect.height() * m_model->songs().size();
	int songAbsTop = rect.height() * indexOfSong;
	float ratio = (double)sb->maximum()/(heightAllItems - area.height());
	int newVal = (int)(ratio * (songAbsTop - (area.height()/2)));
	if (newVal < 1)
		newVal = 0;
	else if (newVal >= sb->maximum())
		newVal = sb->maximum()-1;

	sb->setValue(newVal);
	m_hidden = false;
}

void PlaylistView::savePlaylist() {
	bool ok;
	QString plName = QInputDialog::getText(this, tr("Save playlist as..."),
	                                       tr("Playlist name:"), QLineEdit::Normal, "", &ok);
	while (ok && !plName.isEmpty()) {
		if (!m_mpd->playlistExists(plName)) {
			m_mpd->savePlaylist(plName);
			break;
		}
		QString msg = tr("A playlist with that name already exists.\nPlease use another name:");
		plName = QInputDialog::getText(this, tr("Save playlist as..."),
		                               msg, QLineEdit::Normal, "", &ok);
	}
}

void PlaylistView::jumpToSong() {
	JumpToSongDialog *search = new JumpToSongDialog(this, m_model->songs(), m_model->playingId());
	int id;
	if ((id = search->exec()) > 0) {
		m_mpd->jumpPlaylistId(id);
		scrollTo(m_model->indexId(id));
	}
	delete search;
}

void PlaylistView::information() {
	new MetaInfoDialog(this, m_model->songs(selectedIndexes()));
}

// DnD methods
void PlaylistView::dragEnterEvent(QDragEnterEvent *event) {
	// accept uris
	if (hasDroppableFormat(event->mimeData()))
		event->accept();
}

void PlaylistView::dropEvent(QDropEvent *event) {
	if (m_dropLine.isNull() || m_dropPos == DONTDROP) {
		setDirtyRegion(visibleRegion());
		repaint();
		return;
	}

	const QMimeData *m = event->mimeData();
	if (m->hasFormat("qmpdclient/directory")) {
		event->accept();
		QList<MPDDirectory*> dirs;
		foreach (QString url, QString::fromUtf8(m->data("qmpdclient/directory")).split("\n", QString::SkipEmptyParts)) {
			dirs << new MPDDirectory(url);
		}
		m_mpd->addDirectories(dirs, true, m_dropPos);
		while (!dirs.isEmpty())
			delete dirs.takeLast();
	} else if (m->hasFormat("qmpdclient/song")) {
		event->accept();
		MPDSongList songs;
		foreach (QString url, QString::fromUtf8(m->data("qmpdclient/song")).split("\n", QString::SkipEmptyParts)) {
			songs += MPDSong::createTest(url);
		}
		m_mpd->addSongs(songs, true, m_dropPos);
		while (!songs.isEmpty())
			delete songs.takeFirst();
	} else if (m->hasFormat("qmpdclient/playlist")) {
		event->accept();
		MPDSongList playlists;
		foreach (QString url, QString::fromUtf8(m->data("qmpdclient/playlist")).split("\n", QString::SkipEmptyParts)) {
			playlists += MPDSong::createPlaylist(url);
		}
		m_mpd->addSongs(playlists, true, m_dropPos);
		while (!playlists.isEmpty())
			delete playlists.takeFirst();
	} else if (m->hasFormat("qmpdclient/move")) {
		event->accept();
		// drop pos special for moves
		if (m_dropPos == DROPEND)
			m_dropPos = m_model->lastSong()->pos();
		m_mpd->moveSongs(m_model->songs(selectedIndexes()), m_dropPos);
	}

	// Remove dropline
	m_dropLine = QLine();
	setDirtyRegion(visibleRegion());
	repaint();
}

void PlaylistView::dragLeaveEvent(QDragLeaveEvent *) {
	m_dropPos = DONTDROP;
	m_dropLine = QLine();
	setDirtyRegion(visibleRegion());
	repaint();
}

void PlaylistView::dragMoveEvent(QDragMoveEvent *event) {
	//QTreeView::dragMoveEvent(event);
	int originPos = -1;
	// Check for a move..
	const QMimeData *m = event->mimeData();

	if (!hasDroppableFormat(m))
		return;

	startAutoScroll();

	if (m->hasFormat("qmpdclient/move")) {
		// Find where the songs are being moved from. Need to check for upwards moves.
		originPos = QString::fromUtf8(m->data("qmpdclient/move")).split("\n")[0].toInt();
	}

	QModelIndex mouseIndex = indexAt(event->pos());
	if (!mouseIndex.isValid()) {
		// Index is invalid, must be we are below list.
		MPDSong *song = m_model->lastSong();
		if (!song)
			mouseIndex = QModelIndex();
		else
			mouseIndex = m_model->indexId(song->id());
	}

	QRect itemRect = visualRect(mouseIndex.sibling(mouseIndex.row(), mouseIndex.column()));

	MPDSongList songs = m_model->songs(mouseIndex);
	MPDSong *song = 0;
	if (!songs.isEmpty())
		song = songs.first();

	int y = itemRect.bottom();
	int halfy = itemRect.y() + itemRect.height()/2;
	// Set drop point
	song ? m_dropPos = song->pos() : 0;
	if (!mouseIndex.isValid() || (song == m_model->songs().first() && event->pos().y() < halfy)) {
		// We are at first song, and halfway above item, draw line on top
		y = itemRect.top();
		// Drop at zero, all the way to the top
		m_dropPos = 0;
		// For moves we need the pos above
	} else if (song && song->pos() >= m_model->lastSong()->pos()) {
		// Drop below everything
		m_dropPos = DROPEND;
	} else if (originPos > m_dropPos || originPos == -1) {
		// We are moving upwards, or adding, ++ pos.
		m_dropPos++;
	}
	m_dropLine = QLine(0, y, viewport()->width(), y);
	setDirtyRegion(visibleRegion());
	repaint();
}

bool PlaylistView::hasDroppableFormat(const QMimeData *m) {
	return (m->hasFormat("qmpdclient/song") ||
	        m->hasFormat("qmpdclient/playlist") ||
	        m->hasFormat("qmpdclient/move") ||
	        m->hasFormat("qmpdclient/directory"));
}

void PlaylistView::drawRow(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const {
	QTreeView::drawRow(painter, option, index);
	if (!m_dropLine.isNull())
		painter->drawLine(m_dropLine);
}

void PlaylistView::startDrag(Qt::DropActions) {
	if (m_midButtonDrag)
		return;
	QDrag *drag = new QDrag(this);
	drag->setPixmap(QPixmap(":/icons/crystal/goto.png"));
	QString poses = "";
	foreach (MPDSong *song, m_model->songs(selectedIndexes())) {
		poses += QString::number(song->pos()) + "\n";
	}
	if (poses.isEmpty())
		return;
	QMimeData *mimeData = new QMimeData();
	mimeData->setData("qmpdclient/move", poses.toUtf8());
	drag->setMimeData(mimeData);
	drag->start();
}

==> qmpdclient-1.0.7.tar.gz: qmpdclient-1.0.7/src/playlistview.cpp <==
/*
 * QMPDClient - An MPD client written in Qt 4.
 * Copyright (C) 2005-2006 Hvard Tautra Knutsen <havtknut@tihlde.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"
#include "jumptosongdialog.h"
#include "metainfodialog.h"
#include "mpd.h"
#include "playlistitemdelegate.h"
#include "playlistmodel.h"
#include "playlistview.h"
#include <QContextMenuEvent>
#include <QInputDialog>
#include <QMenu>
#include <QPainter>

PlaylistView::PlaylistView(QWidget *parent) : QListView(parent),
		m_config(Config::instance()),
		m_mpd(MPD::instance()),
m_menu(new QMenu(this)) {
	setItemDelegate(new PlaylistDelegate(this));
	setModel(m_model = new PlaylistModel(this));
	m_model->setPattern(m_config->playlistPattern());

	// playlist menu
	QAction *jumpToAction = m_menu->addAction(QPixmap(":/icons/crystal/goto.png"), tr("J&ump to song..."), this, SLOT(jumpToSong()), tr("Ctrl+U", "This is the shortcut for 'Jump to song in playlist'"));

	addAction(jumpToAction);
	m_menu->addSeparator();
	m_removeAction = m_menu->addAction(QPixmap(":/icons/remove.png"), tr("&Remove"), this, SLOT(removeItems()));
	m_removeAction->setShortcut(Qt::Key_Delete);
	m_removeAction->setShortcutContext(Qt::WidgetShortcut);
	addAction(m_removeAction);
	m_cropAction = m_menu->addAction(QPixmap(":/icons/crop.png"), tr("&Crop playlist"), this, SLOT(cropItems()));
	m_menu->addAction(QPixmap(":/icons/removeall.png"), tr("Remove &all"), m_mpd, SLOT(clearPlaylist()));
	m_menu->addSeparator();
	m_randomAction = m_menu->addAction(QPixmap(":/icons/amarok/random.png"), tr("Ra&ndom play"));
	m_repeatAction = m_menu->addAction(QPixmap(":/icons/amarok/repeat.png"), tr("Re&peat play"));
	m_menu->addAction(QPixmap(":/icons/amarok/shuffle.png"), tr("S&huffle playlist"), m_mpd, SLOT(shufflePlaylist()));
	m_menu->addSeparator();
	m_menu->addAction(QPixmap(":/icons/crystal/saveas.png"), tr("&Save playlist..."), this, SLOT(savePlaylist()));
	m_informationAction = m_menu->addAction(QPixmap(":/icons/crystal/info.png"), tr("&Information..."), this, SLOT(information()));
	m_randomAction->setCheckable(true);
	m_repeatAction->setCheckable(true);

	// signals and slots connections
	connect(m_mpd, SIGNAL(playingSongUpdated(MPDSong)), this, SLOT(setSong(MPDSong)));
	connect(m_mpd, SIGNAL(playlistUpdated(MPDSongList)), this, SLOT(setPlaylist(MPDSongList)));
	connect(m_mpd, SIGNAL(randomUpdated(bool)), m_randomAction, SLOT(setChecked(bool)));
	connect(m_mpd, SIGNAL(repeatUpdated(bool)), m_repeatAction, SLOT(setChecked(bool)));
	connect(m_randomAction, SIGNAL(toggled(bool)), m_mpd, SLOT(setRandom(bool)));
	connect(m_repeatAction, SIGNAL(toggled(bool)), m_mpd, SLOT(setRepeat(bool)));
	connect(this, SIGNAL(activated(const QModelIndex &)), this, SLOT(indexActivated(const QModelIndex &)));

	selectionChanged();
}

void PlaylistView::selectionChanged(const QItemSelection &s, const QItemSelection &u) {
	QListView::selectionChanged(s, u);
	bool t = selectedIndexes().count() > 0;
	m_cropAction->setEnabled(t);
	m_informationAction->setEnabled(t);
	m_removeAction->setEnabled(t);
	emit toggleActions(t);
}


void PlaylistView::indexActivated(const QModelIndex &index) {
	foreach(MPDSong s, m_model->songs(index)) {
		m_mpd->jumpPlaylistSong(s);
		break;
	}
}

void PlaylistView::cropItems() {
	m_mpd->removeSongs(m_model->unSelectedSongs(selectedIndexes()));
}

void PlaylistView::removeItems() {
	QModelIndexList selected = selectedIndexes();
	if (selected.isEmpty())
		return;
	m_mpd->removeSongs(m_model->songs(selected));
}

void PlaylistView::setPlaylist(const MPDSongList &songs) {
	QModelIndexList selected = selectedIndexes();
	QModelIndex last = selected.isEmpty() ? QModelIndex() : selected.last();
	m_model->setSongs(songs);
	setCurrentIndex(last);
	selectionChanged();
}

void PlaylistView::setSong(const MPDSong &s) {
	m_model->setPlaying(s);
	if (!s.isNull() && m_config->scrollToPlaying()) {
		scrollTo(m_model->indexOfSong(s), QAbstractItemView::PositionAtCenter);
		m_hidden = true;
	}
}

void PlaylistView::hideEvent(QHideEvent *e) {
	QListView::hideEvent(e);
	m_hidden = true;
}
void PlaylistView::paintEvent(QPaintEvent *e) {
	QListView::paintEvent(e);
	if (!m_dropLine.isNull()) {
		QPainter painter(viewport());
		painter.drawLine(m_dropLine);
	}
	if (m_hidden && m_config->scrollToPlaying()) {
		scrollTo(m_model->indexOfSong(m_model->playingSong()), QAbstractItemView::PositionAtCenter);
		m_hidden = false;
	}
}

void PlaylistView::savePlaylist() {
	bool ok;
	QString plName = QInputDialog::getText(this, tr("Save playlist as..."),
	                                       tr("Playlist name:"), QLineEdit::Normal, "", &ok);
	while (ok && !plName.isEmpty()) {
		if (!m_mpd->playlistExists(plName)) {
			m_mpd->savePlaylist(plName);
			break;
		}
		QString msg = tr("A playlist with that name already exists.\nPlease use another name:");
		plName = QInputDialog::getText(this, tr("Save playlist as..."),
		                               msg, QLineEdit::Normal, "", &ok);
	}
}

void PlaylistView::jumpToSong() {
	JumpToSongDialog *search = new JumpToSongDialog(this, m_model->songs(), m_model->playingSong());
	if (search->exec() == QDialog::Accepted) {
		m_mpd->jumpPlaylistSong(search->jumpSong());
		if (m_config->scrollToPlaying())
			scrollTo(m_model->indexOfSong(search->jumpSong()), QAbstractItemView::PositionAtCenter);
	}
	delete search;
}

void PlaylistView::information() {
	new MetaInfoDialog(this, m_model->songs(selectedIndexes()));
}

void PlaylistView::contextMenuEvent(QContextMenuEvent *e) {
	QModelIndex clicked = indexAt(e->pos());
	QModelIndexList indexes = selectedIndexes();
	if (indexes.isEmpty() || !indexes.contains(clicked))
		setCurrentIndex(indexAt(e->pos()));
	m_menu->popup(e->globalPos());
}

// Hacks to enable contigous selections with middle mouse button.
void PlaylistView::mouseMoveEvent(QMouseEvent *e) {
	QMouseEvent *e2 = e;
	if (m_model->isDragDisabled())
		e2 = new QMouseEvent(e->type(), e->pos(), Qt::LeftButton, e->buttons() | Qt::LeftButton, e->modifiers());
	QListView::mouseMoveEvent(e2);
}

void PlaylistView::mousePressEvent(QMouseEvent *e) {
	QMouseEvent *e2 = e;
	if (e->button() == Qt::MidButton) {
		m_model->setDragDisabled(true);
		e2 = new QMouseEvent(e->type(), e->pos(), Qt::LeftButton, e->buttons(), e->modifiers());
	}
	QListView::mousePressEvent(e2);
}

void PlaylistView::mouseReleaseEvent(QMouseEvent *e) {
	if (e->button() == Qt::MidButton)
		m_model->setDragDisabled(false);
	QListView::mouseReleaseEvent(e);
}

// Set up custom drop indicator, since the default one does not behave as I wish.
void PlaylistView::dragMoveEvent(QDragMoveEvent *e) {
	QListView::dragMoveEvent(e);
	if (m_model->rowCount() < 1) // The list is empty, draw nothing
		return;

	startAutoScroll();

	QModelIndex mouseIndex = indexAt(e->pos());
	if (!mouseIndex.isValid())
		mouseIndex = m_model->index(m_model->rowCount()-1, 0);

	QRect itemRect = visualRect(mouseIndex);

	int y = itemRect.bottom();
	int halfy = itemRect.y() + itemRect.height()/2;

	if (mouseIndex.row() == 0 && e->pos().y() < halfy) {
		// We are at first song, and halfway above item, draw line on top
		y = itemRect.top();
		m_model->setTopDrop(true);
	} else
		m_model->setTopDrop(false);

	m_dropLine = QLine(0, y, viewport()->width(), y);
	setDirtyRegion(visibleRegion());
	repaint();
}

// Remove drop indicator
void PlaylistView::dropEvent(QDropEvent *e) {
	QListView::dropEvent(e);
	m_dropLine = QLine();
	setDirtyRegion(visibleRegion());
	repaint();
}

// Remove drop indicator
void PlaylistView::dragLeaveEvent(QDragLeaveEvent *e) {
	QListView::dragLeaveEvent(e);
	m_dropLine = QLine();
	setDirtyRegion(visibleRegion());
	repaint();
}
